<?php

namespace Shavshukov\EventAppBase;

use Shavshukov\EventAppBase\FileReaders\FileReaderInterface;
use Shavshukov\EventAppBase\Interfaces\ModelInterface;
use Shavshukov\EventAppBase\Service\ErrorReport;
use Shavshukov\EventAppBase\Service\Container;

use React\EventLoop\Factory as LoopFactory;
use React\EventLoop\LoopInterface;

abstract class Application extends Container implements ApplicationInterface
{
    /**
     * Application event loop - react loop interface instance
     *
     * @var \React\EventLoop\LoopInterface;
     */
    protected $loop;
    
    /**
     * Configuration Model Instance
     *
     * @var \Shavshukov\EventAppBase\Interfaces\ModelInterface
     */
    protected $config;
    
    /**
     * File reader instancc
     *
     * @var \Shavshukov\EventAppBase\FileReaders\FileReaderInterface
     */
    protected $fileReader;
    
    /**
     * Path to the application
     *
     * @var string
     */
    protected $appPath;
    
    /**
     * Constructor
     * 
     * @param string $appPath
     */
    public function __construct(string $appPath)
    {
        ErrorReport::setInstance(new ErrorReport($this));
        
        $this->loop = LoopFactory::create();
        $this->appPath = rtrim($appPath, DIRECTORY_SEPARATOR);
        $this->fileReader = $this->getFileReaderInstance();
        $this->config = $this->getConfigInstance();
        
        // bind an application and a loop instances to the service provider self as services
        $this->bind('app', $this, true);
    }
    
    /**
     * Get new config model instance
     * 
     * @return \Shavshukov\EventAppBase\Interfaces\ModelInterface
     */
    protected abstract function getConfigInstance(): ModelInterface;
    
    /**
     * Get new FileReaderInterface implementation based on driver from config
     * 
     * @return \Shavshukov\EventAppBase\FileReaders\FileReaderInterface
     * @throws \Exception
     */
    protected function getFileReaderInstance(): FileReaderInterface
    {
        // Under windows, files cannot be read in a non-blocking way
        $isWindows = stripos(PHP_OS, 'WIN') === 0;
        $readerDriver = $isWindows ? 'blocking' : 'stream';
        
        switch ($readerDriver) {
            case 'blocking': return new \Shavshukov\EventAppBase\FileReaders\BlockingFileReader();
            case 'stream': return new \Shavshukov\EventAppBase\FileReaders\StreamedFileReader($this->getLoop());
            default: throw new \Exception("Unsupported file_reader_driver: {$readerDriver}");
        }
    }
    
    /**
     * {@inheritDoc}
     */    
    public function getAppPath(): string
    {
        return $this->appPath;
    }
    
    /**
     * {@inheritDoc}
     */    
    public function getLoop(): LoopInterface
    {
        return $this->loop;
    }
    
    /**
     * {@inheritDoc}
     */    
    public function getFileReader(): FileReaderInterface
    {
        return $this->fileReader;
    }
    
    /**
     * {@inheritDoc}
     */    
    public function config(string $configPath = null, $defaultValue = null)
    {
        if ($configPath === null) {
            return $this->config;
        }
        
        return $this->config->get($configPath, $defaultValue);
    }
    
    /**
     * {@inheritDoc}
     */    
    public function service(string $serviceName = null, ...$arguments)
    {
        if ($serviceName === null) {
            return $this;
        }
        
        return $this->resolve($serviceName, ...$arguments);
    }
    
    /**
     * {@inheritDoc}
     */    
    public function run()
    {
        $this->getLoop()->run();
    }
    
    /**
     * {@inheritDoc}
     */    
    public function stop()
    {
        if (!isset($this->loop)) {
            exit(-1);
        }
        
        $this->getLoop()->stop();
    }
    
}