<?php

namespace Shavshukov\EventAppBase\Http\Routing;

use Shavshukov\EventAppBase\Service\ErrorReport;
use Shavshukov\EventAppBase\ApplicationInterface;
use Shavshukov\EventAppBase\Http\Controller;
use Shavshukov\EventAppBase\Http\Server;

use Psr\Http\Message\ServerRequestInterface;
use React\Promise;
use React\Http\Response;

abstract class Router
{
    /**
     * Application instance
     *
     * @var \Shavshukov\EventAppBase\ApplicationInterface;
     */
    protected $app;
    
    /**
     * Server instance
     *
     * @var \Shavshukov\EventAppBase\Http\Server
     */
    protected $server;
    
    /**
     * Array of routes
     *
     * @var array
     */
    protected $routes;
    
    /**
     * Constructor
     * 
     * @param \Shavshukov\EventAppBase\Http\Server $server
     * @param \Shavshukov\EventAppBase\ApplicationInterface $app
     */
    public function __construct(Server $server, ApplicationInterface $app)
    {
        $this->app = $app;
        $this->server = $server;
        $this->routes = $this->getRoutes();
        
        if (!isset($this->routes['/403'], $this->routes['/404'])) {
            throw new \Exception("Router must provide necessary routes for errors: '/403', '/404'");
        }
    }
    
    /**
     * Method to get routes enumeration
     * 
     * @return array
     */
    abstract public function getRoutes(): array;
    
    /**
     * Whether the client have authorization to the routes proviced by this Router and provide redirection if necessary
     * 
     * @param string $method
     * @param string $path
     * @return \React\Http\Promise\PromiseInterface
     */
    abstract public function authorise(ServerRequestInterface $request, string $method, string $path): Promise\PromiseInterface;
    
    /**
     * Resolve controller by name
     * 
     * @param string $controllerName
     * @return \Shavshukov\EventAppBase\Http\Controller
     */
    abstract public function resolveController(string $controllerName): Controller;
    
    /**
     * Find route for request and run its controller's action or return a response
     * 
     * @param \Psr\Http\Message\ServerRequestInterface $request
     * @return \React\Http\Promise\PromiseInterface<\React\Http\Response>
     * @throws \Exception
     */
    public function processRequest(ServerRequestInterface $request): Promise\PromiseInterface
    {
        return new Promise\Promise(function ($resolver) use ($request) {
            $matcher = new Matcher($this->app->getFileReader(), $this, $request);
            $matcher->match()->then(function ($matched) use ($request) {
                if (is_array($matched)) {
                    list($controllerName, $action) = array_values($matched);
                    $controller = $this->resolveController($controllerName);
                    return $controller->{$action}($request);
                } else {
                    return $matched;
                }
            })->then(function ($result) use ($resolver) {
                $resolver($result);
            }, function (\Throwable $e) use ($resolver) {
                $report = ErrorReport::report($e, false);
                $resolver(new Response(500, ['Content-Type' => 'text/plain'], $report));
            });
        });
    }
    
    /**
     * Get a route for path and method
     * 
     * @param string $path
     * @param string $method
     * @return string|array|null
     */
    public function getRoute(string $path, string $method)
    {
        if (!empty($this->routes[$path])) {
            $route = $this->routes[$path];
            if (is_string($route)) {
                return $route;
            }
            
            return $route[$method] ?? null;
        }
    }
    
    /**
     * Get real path of public folder
     * 
     * @return string
     */
    public function getPublicRealPath(): string
    {
        return $this->server->getPublicRealPath();
    }
}
