<?php

namespace Shavshukov\EventAppBase\Service;

use Shavshukov\EventAppBase\ApplicationInterface;

class ErrorReport
{
    /**
     * Indent size for each part of the block of error
     * 
     * @var int
     */
    public static $indent = 2;
    
    /**
     * ErrorReport instance
     *
     * @var \Shavshukov\EventAppBase\Service\ErrorReport
     */
    protected static $instance;
    
    /**
     * Application instance
     *
     * @var \Shavshukov\EventAppBase\ApplicationInterface
     */
    protected $app;
    
    /**
     * Constructor
     * 
     * @param \Shavshukov\EventAppBase\ApplicationInterface $app
     */
    public function __construct(ApplicationInterface $app)
    {
        $this->app = $app;
    }
    
    /**
     * Set global instance to a given instance, possibly a sub-class
     * 
     * @param \Shavshukov\EventAppBase\Service\ErrorReport $instance
     */
    public static function setInstance(ErrorReport $instance)
    {
        self::$instance = $instance;
    }
    
    /**
     * Report an error gracefully
     * 
     * @param \Throwable $e
     * @return int
     */
    public function reportError(\Throwable $e, bool $critical)
    {
        $formatted = static::formatErrorBlocks($e);
        echo $formatted;
        
        if ($critical) {
            $this->app->stop();
        }
        
        return $formatted;
    }
    
    /**
     * Report an error
     * 
     * @param \Throwable $e
     * @return string
     */
    public static function report(\Throwable $e, bool $critical)
    {
        if (isset(self::$instance)) {
            return self::$instance->reportError($e, $critical);
        } else {
            // fallback to printing the info and exiting
            $formatted = static::formatErrorBlocks($e);
            fwrite(STDERR, $formatted);
            
            if ($critical) {
                exit(-1);
            }
            
            return $formatted;
        }
    }
    
    /**
     * Get error description and trace blocks for error and any number of its previous
     * 
     * @param \Throwable $e
     * @retrun array
     */
    public static function getErrorInfo(\Throwable $e): array
    {
        $blocks = [];
        $meet = [];
        
        do {
            $lines = [
                'error' => get_class($e),
                'data' => [
                    'code' => $e->getCode(),
                    'message' => $e->getMessage(),
                    'location' => "{$e->getFile()}:{$e->getLine()}",
                ],
                'trace' => $e->getTraceAsString()
            ];
            
            $previous = $e->getPrevious();
            
            if ($e === $previous) {
                $e = null;
            } else {
                $e = $previous;
            }
            
            $blocks[] = $lines;
        } while ($e !== null && !in_array($e, $meet));
        
        return $blocks;
    }
    
    /**
     * Format error description and trace blocks for error and any number of its previous
     * 
     * @param \Throwable $e
     * @return string
     */
    public static function formatErrorBlocks(\Throwable $e)
    {
        $blocks = static::getErrorInfo($e);
        $blocksStrings = [];
        $blockLineIndent = str_repeat(' ', static::$indent);
        $eol = PHP_EOL;
        
        foreach ($blocks as $block) {
            $blockLinesStrings = [];
            $blockLinesStrings[] = "[{$block['error']}]{$eol}[";
            foreach ($block['data'] as $key => $value) {
                if ($value !== null && $value !== '') {
                    $key = ucfirst($key);
                    $blockLinesStrings[] = "{$blockLineIndent}{$key}: {$value}";
                }
            }
            $blockLinesStrings[] = ']';
            $blockLinesStrings[] = $block['trace'];
            $blocksStrings[] = implode($eol, $blockLinesStrings);
        }
        
        $blocksDelimeter = "{$eol}{$eol}Previous:{$eol}";
        $implodeBlocksStrings = implode($blocksDelimeter, $blocksStrings);
        return $implodeBlocksStrings . $eol;
    }
}
